
/* 
 * odbctcl.cpp --
 *
 *	This file provides ODBC services for TCL applications
 *
 * Copyright (c) 1996-1997 Jos L Porcayo.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include "tcl.h"

#ifdef __WIN32__
  #include <windows.h>
#else
  #include <stdlib.h>
  #define wsprintf sprintf
  #define _export
#endif

#include "sql.h"
#include "sqlext.h"
#include <string.h>

//
// Global variables (henv) & type definitions
//

HENV henv = 0;
typedef struct {
  HDBC hdbc;
  HSTMT hstmt;
  char *result;
} state, *pstate;

char * def_result_var = "sqlres";
char *api_fn[100];

#define STATE pstate(clientData)
#define VERIFY(flags) if (rc != SQL_SUCCESS && rc != SQL_SUCCESS_WITH_INFO) throw(sqlError(flags))
#define CLOSE 0x1
#define RESETP 0x2
#define NULLHSTMT 0x4
#define NULLHDBC 0x8
#define SPARAM(index) (UCHAR *) (argv[index][0] ? argv[index] : NULL), SQL_NTS

class sqlError {
  int flags;
public:
  sqlError(int Flags = 0) {flags = Flags;}
  inline int get(void) {return flags;}
  int explain(Tcl_Interp *interp, pstate st);
};

class argsError {
  char *msg;
public:
  argsError(char *Msg) {msg = Msg;}
  inline char *get(void) {return msg;}
  int explain(Tcl_Interp *interp, char *command);
};

class msgError {
  char *msg;
public:
  msgError(char *Msg) {msg = Msg;}
  inline char *get(void) {return msg;}
  int explain(Tcl_Interp *interp);
};

#define CONNECT_OPTIONS 5
struct {
  char *name;
  UWORD identifier;
} connect_option[CONNECT_OPTIONS] = { {"-accessmode", SQL_ACCESS_MODE},
    {"-autocommit", SQL_AUTOCOMMIT}, {"-currentqualifier", SQL_CURRENT_QUALIFIER},
    {"-opttrace", SQL_OPT_TRACE}, {"-opttracefile", SQL_OPT_TRACEFILE} };

class connOptsError {
  char *msg;
public:
  connOptsError(char *Msg) {msg = Msg;}
  int explain(Tcl_Interp *interp);
};

class getinfoError {
  char *msg;
public:
  getinfoError(char *Msg) {msg = Msg;}
  int explain(Tcl_Interp *interp);
};

//
// initialize: initialize translation tables to map some ODBC defines to
//             human-readable form.
//
void initialize(void) {
  /* initialize exists array with descriptions for ODBC functions */
  /* Core Functions */
  api_fn[SQL_API_SQLALLOCCONNECT]     = "allocconnect";
  api_fn[SQL_API_SQLALLOCENV]         = "allocenv";
  api_fn[SQL_API_SQLALLOCSTMT]        = "allocstmt";
  api_fn[SQL_API_SQLBINDCOL]          = "bindcol";
  api_fn[SQL_API_SQLCANCEL]           = "cancel";
  api_fn[SQL_API_SQLCOLATTRIBUTES]    = "colattributes";
  api_fn[SQL_API_SQLCONNECT]          = "connect";
  api_fn[SQL_API_SQLDESCRIBECOL]      = "describecol";
  api_fn[SQL_API_SQLDISCONNECT]       = "disconnect";
  api_fn[SQL_API_SQLERROR]            = "error";
  api_fn[SQL_API_SQLEXECDIRECT]       = "execdirect";
  api_fn[SQL_API_SQLEXECUTE]          = "execute";
  api_fn[SQL_API_SQLFETCH]            = "fetch";
  api_fn[SQL_API_SQLFREECONNECT]      = "freeconnect";
  api_fn[SQL_API_SQLFREEENV]          = "freeenv";
  api_fn[SQL_API_SQLFREESTMT]         = "freestmt";
  api_fn[SQL_API_SQLGETCURSORNAME]    = "getcursorname";
  api_fn[SQL_API_SQLNUMRESULTCOLS]    = "numresultcols";
  api_fn[SQL_API_SQLPREPARE]          = "prepare";
  api_fn[SQL_API_SQLROWCOUNT]         = "rowcount";
  api_fn[SQL_API_SQLSETCURSORNAME]    = "setcursorname";
  api_fn[SQL_API_SQLSETPARAM]         = "setparam";
  api_fn[SQL_API_SQLTRANSACT]         = "transact";
  /* Level 1 Functions */
  api_fn[SQL_API_SQLCOLUMNS]          = "columns";
  api_fn[SQL_API_SQLDRIVERCONNECT]    = "driverconnect";
  api_fn[SQL_API_SQLGETCONNECTOPTION] = "getconnectoption";
  api_fn[SQL_API_SQLGETDATA]          = "getdata";
  api_fn[SQL_API_SQLGETFUNCTIONS]     = "getfunctions";
  api_fn[SQL_API_SQLGETINFO]          = "getinfo";
  api_fn[SQL_API_SQLGETSTMTOPTION]    = "getstmtoption";
  api_fn[SQL_API_SQLGETTYPEINFO]      = "gettypeinfo";
  api_fn[SQL_API_SQLPARAMDATA]        = "paramdata";
  api_fn[SQL_API_SQLPUTDATA]          = "putdata";
  api_fn[SQL_API_SQLSETCONNECTOPTION] = "setconnectoption";
  api_fn[SQL_API_SQLSETSTMTOPTION]    = "setstmtoption";
  api_fn[SQL_API_SQLSPECIALCOLUMNS]   = "specialcolumns";
  api_fn[SQL_API_SQLSTATISTICS]       = "statistics";
  api_fn[SQL_API_SQLTABLES]           = "tables";
  /* Level 2 Functions */
  api_fn[SQL_API_SQLBROWSECONNECT]    = "browseconnect";
  api_fn[SQL_API_SQLCOLUMNPRIVILEGES] = "columnprivileges";
  api_fn[SQL_API_SQLDATASOURCES]      = "datasources";
  api_fn[SQL_API_SQLDESCRIBEPARAM]    = "describeparam";
  api_fn[SQL_API_SQLEXTENDEDFETCH]    = "extendedfetch";
  api_fn[SQL_API_SQLFOREIGNKEYS]      = "foreignkeys";
  api_fn[SQL_API_SQLMORERESULTS]      = "moreresults";
  api_fn[SQL_API_SQLNATIVESQL]        = "nativesql";
  api_fn[SQL_API_SQLNUMPARAMS]        = "numparams";
  api_fn[SQL_API_SQLPARAMOPTIONS]     = "paramoptions";
  api_fn[SQL_API_SQLPRIMARYKEYS]      = "primarykeys";
  api_fn[SQL_API_SQLPROCEDURECOLUMNS] = "procedurecolumns";
  api_fn[SQL_API_SQLPROCEDURES]       = "procedures";
  api_fn[SQL_API_SQLSETPOS]           = "setpos";
  api_fn[SQL_API_SQLSETSCROLLOPTIONS] = "setscrolloptions";
  api_fn[SQL_API_SQLTABLEPRIVILEGES]  = "tableprivileges";
  /* SDK 2.0 Additions */
  api_fn[SQL_API_SQLDRIVERS]          = "drivers";
  api_fn[SQL_API_SQLBINDPARAMETER]    = "bindparameter";
}

#ifdef __WIN32__
//
// Win32 DLL initialization
//
BOOL WINAPI DllEntryPoint(HINSTANCE, DWORD fdwReason, LPVOID)
{
  switch (fdwReason) {
    case DLL_PROCESS_ATTACH: break;
    case DLL_PROCESS_DETACH: SQLFreeEnv(henv); break;
    default: break;
  }
  return TRUE;
}
#endif

//
// explain: return error information to the interpreter
//
int sqlError::explain(Tcl_Interp *interp, pstate st)
{
  SDWORD native_err;
  SWORD  cb;
  UCHAR  sqlstate[10];
  UCHAR  msg[SQL_MAX_MESSAGE_LENGTH-1];

  SQLError(henv, st ? (flags & NULLHDBC ? SQL_NULL_HDBC : st->hdbc) : SQL_NULL_HDBC,
        st ? (flags & NULLHSTMT ? SQL_NULL_HSTMT : st->hstmt) : SQL_NULL_HSTMT,
        sqlstate, &native_err, msg, sizeof(msg), &cb);
  Tcl_SetResult(interp, (char *) msg, TCL_VOLATILE);
  Tcl_SetErrorCode(interp, (char *) sqlstate, NULL);
  if (flags & CLOSE) { SQLFreeStmt(st->hstmt, SQL_CLOSE); }
  if (flags & RESETP) { SQLFreeStmt(st->hstmt, SQL_RESET_PARAMS); }
  return TCL_ERROR;
}

int argsError::explain(Tcl_Interp *interp, char *command)
{
  Tcl_AppendResult(interp, "wrong # args: should be \"", command, " ", msg, "\"", NULL);
  return TCL_ERROR;
}

int msgError::explain(Tcl_Interp *interp)
{
  Tcl_AppendResult(interp, msg, NULL);
  return TCL_ERROR;
}

int connOptsError::explain(Tcl_Interp *interp) {
  Tcl_ResetResult(interp);
  Tcl_AppendResult(interp, "bad connection option \"", msg, "\", must be: ", NULL);
  int i;
  for (i = 0; i < CONNECT_OPTIONS - 1; i++)
    Tcl_AppendResult(interp, connect_option[i].name, ", ", NULL);
  Tcl_AppendResult(interp, connect_option[i].name, NULL);
  return TCL_ERROR;
}

//
// bind_params: binds parameters for a SQL statement
//
int bind_params(pstate st, int argc, char *argv[])
{
  int reset = 0;
  if (argc) for (int i = 0; i < argc; i++) {
    SDWORD cb = strlen(argv[i]);
    RETCODE rc = SQLBindParameter(st->hstmt, UWORD(i+1), SQL_PARAM_INPUT, SQL_C_CHAR, SQL_CHAR,
                cb, 0, argv[i], 0, &cb);
    VERIFY(RESETP);
    reset++;
  }
  return reset;
}

//
// get_results: retrieves results from any ODBC statement
//
int get_results(pstate st, Tcl_Interp *interp)
{
  SWORD cols; SDWORD rows;
  RETCODE rc = SQLRowCount(st->hstmt, &rows); VERIFY( CLOSE );
  rc = SQLNumResultCols(st->hstmt, &cols); VERIFY( CLOSE );
  if ( ! cols ) { // not a select, so return affected rows (if any, can be -1)
    wsprintf(interp->result, "%ld", rows);
    return TCL_OK;
  }
  // prepare to return values in result array
  Tcl_UnsetVar(interp, st->result, 0);
  int col; char idx[20]; static char buf[2600]; SDWORD cb;
  SWORD wcb, type, scale, nullable; UDWORD precision;
  SWORD types[200];
  for (col = 1; col <= cols; col++) {
    wsprintf(idx, "0,%d", col);
    rc = SQLDescribeCol(st->hstmt, (UWORD) col, (UCHAR *) buf, sizeof(buf), &wcb,
		  &type, &precision, &scale, &nullable); VERIFY( CLOSE );
    types[col] = type;
    Tcl_SetVar2(interp, st->result, idx, buf, 0);
    wsprintf(buf, " %d %ld %d %d", type, precision, scale, nullable);
    Tcl_SetVar2(interp, st->result, idx, buf, TCL_APPEND_VALUE);
  }
  int row = 1; rc = SQLFetch(st->hstmt);
  while ( rc == SQL_SUCCESS ) {
    for (col = 1; col <= cols; col++) {
      wsprintf(idx, "%d,%d", row, col);
      if ( types[col] == SQL_TIME || types[col] == SQL_DATE || types[col] == SQL_TIMESTAMP ) {
        TIMESTAMP_STRUCT ts;
        rc = SQLGetData(st->hstmt, (UWORD) col, SQL_C_TIMESTAMP, &ts, sizeof(ts), &cb);
        VERIFY( CLOSE );
        if ( cb == SQL_NULL_DATA )
          Tcl_SetVar2(interp, st->result, idx, "null", 0);
        else {
          wsprintf(buf, "{ts '%4d-%02d-%02d %02d:%02d:%02d'}", ts.year, ts.month, ts.day,
                ts.hour, ts.minute, ts.second);
          Tcl_SetVar2(interp, st->result, idx, buf, 0);
        }
      } else {
        rc = SQLGetData(st->hstmt, (UWORD) col, SQL_C_CHAR, buf, sizeof(buf), &cb);
        VERIFY( CLOSE );
        if ( cb == SQL_NULL_DATA )
          Tcl_SetVar2(interp, st->result, idx, "null", 0);
        else
          Tcl_SetVar2(interp, st->result, idx, buf, 0);
      }
    }
    row++; rc = SQLFetch(st->hstmt);
  }
  wsprintf(interp->result, "%ld %ld", row - 1, cols);
  SQLFreeStmt(st->hstmt, SQL_CLOSE);
  return TCL_OK;
}

//
// option_identifier: receives option name & returns its numeric equivalent
//
UWORD option_identifier(char *option)
{
  for (int i = 0; i < CONNECT_OPTIONS; i++)
    if ( ! strcmp(connect_option[i].name, option) )
      return connect_option[i].identifier;
  throw(connOptsError(option));
}

//
// option_name: receives option identifier & returns its description
//
char * option_name(UWORD option)
{
  for (int i = 0; i < CONNECT_OPTIONS; i++)
    if ( connect_option[i].identifier == option )
      return connect_option[i].name;
  throw(msgError("internal error: missing description for an option"));
}

//
// cget: gets value of a connection configuration option
//
void cget(pstate st, UWORD option, char *value)
{
  int value32; RETCODE rc;
  switch (option) {
    case SQL_ACCESS_MODE : rc = SQLGetConnectOption(st->hdbc, SQL_ACCESS_MODE, (PTR) &value32);
             VERIFY( NULLHSTMT );
             strcpy(value, value32 == SQL_MODE_READ_ONLY ? "readonly" :
                  (value32 == SQL_MODE_READ_WRITE ? "readwrite" : "unknown"));
             break;
    case SQL_AUTOCOMMIT  : rc = SQLGetConnectOption(st->hdbc, SQL_AUTOCOMMIT, (PTR) &value32);
             VERIFY( NULLHSTMT );
             strcpy(value, value32 == SQL_AUTOCOMMIT_OFF ? "off" :
                  (value32 == SQL_AUTOCOMMIT_ON ? "on" : "unknown"));
             break;
    case SQL_OPT_TRACE   : rc = SQLGetConnectOption(st->hdbc, SQL_OPT_TRACE, (PTR) &value32);
             VERIFY( NULLHSTMT );
             strcpy(value, value32 == SQL_OPT_TRACE_OFF ? "off" :
                  (value32 == SQL_OPT_TRACE_ON ? "on" : "unknown"));
             break;
    case SQL_CURRENT_QUALIFIER : rc = SQLGetConnectOption(st->hdbc, SQL_CURRENT_QUALIFIER, (PTR) value);
             VERIFY( NULLHSTMT );
             break;
    case SQL_OPT_TRACEFILE : rc = SQLGetConnectOption(st->hdbc, SQL_OPT_TRACEFILE, (PTR) value);
             VERIFY( NULLHSTMT );
             break;
  }
}

//
// config: sets/gets configuration options
//
int config(pstate st, Tcl_Interp *interp, int argc, char *argv[])
{
  RETCODE rc;
  if ( ! argc ) { // return list of values
    char value[SQL_MAX_OPTION_STRING_LENGTH+1];
    for (int i = 0; i < CONNECT_OPTIONS; i++) {
      Tcl_AppendResult(interp, i ? " {" : "{", connect_option[i].name, NULL);
      cget(st, connect_option[i].identifier, value);
      Tcl_AppendElement(interp, value);
      Tcl_AppendResult(interp, "}", NULL);
    }
  } else if ( argc % 2 ) { // odd number of params is not allowed
    throw(argsError("configure ?-option value? ..."));
  } else for (int i = 0; i < argc; i += 2) {
    UWORD option = option_identifier(argv[i]); int value;
    switch (option) {
      case SQL_ACCESS_MODE :
          if ( !strcmp(argv[i+1], "readonly") )
            rc = SQLSetConnectOption(st->hdbc, SQL_ACCESS_MODE, SQL_MODE_READ_ONLY);
          else if ( !strcmp(argv[i+1], "readwrite") )
            rc = SQLSetConnectOption(st->hdbc, SQL_ACCESS_MODE, SQL_MODE_READ_WRITE);
          else
            throw(msgError("bad value for option \"-accessmode\" must be: readonly or readwrite"));
          break;
      case SQL_AUTOCOMMIT :
          if ( Tcl_GetBoolean(interp, argv[i+1], &value) != TCL_OK ) return TCL_ERROR;
          rc = SQLSetConnectOption(st->hdbc, SQL_AUTOCOMMIT, value ? SQL_AUTOCOMMIT_ON : SQL_AUTOCOMMIT_OFF);
          break;
      case SQL_CURRENT_QUALIFIER :
          rc = SQLSetConnectOption(st->hdbc, SQL_CURRENT_QUALIFIER, (UDWORD) argv[i+1]);
          break;
      case SQL_OPT_TRACE :
          if ( Tcl_GetBoolean(interp, argv[i+1], &value) != TCL_OK ) return TCL_ERROR;
          rc = SQLSetConnectOption(st->hdbc, SQL_OPT_TRACE, value ? SQL_OPT_TRACE_ON : SQL_OPT_TRACE_OFF);
          break;
      case SQL_OPT_TRACEFILE :
          rc = SQLSetConnectOption(st->hdbc, SQL_OPT_TRACEFILE, (UDWORD) argv[i+1]);
          break;
    }
    VERIFY( NULLHSTMT );
  }
  return TCL_OK;
}

#define STRING 1
#define INT16  2
#define INT32  3
#define ENUM16 4
#define ENUM32 5
#define MASK16 6
#define MASK32 7

#define XLAT(id) #id, id, sizeof(id) / sizeof(gvals)

struct gvals {
  int value;
  char *label;
};
struct ginfo {
  UWORD infotype;
  int datatype;
  char *label;
  gvals *values;
  int nvalues;
};
ginfo info_dbms[] = { {SQL_DATABASE_NAME, STRING, "databasename"},
  {SQL_DBMS_NAME, STRING, "dbmsname"},
  {SQL_DBMS_VER, STRING, "dbmsver"} };

gvals fetchdirection[] = { {SQL_FD_FETCH_NEXT, "next"},
  {SQL_FD_FETCH_FIRST, "first"},
  {SQL_FD_FETCH_LAST, "last"},
  {SQL_FD_FETCH_PRIOR, "prior"},
  {SQL_FD_FETCH_ABSOLUTE, "absolute"},
  {SQL_FD_FETCH_RELATIVE, "relative"},
  {SQL_FD_FETCH_RESUME, "resume"},
  {SQL_FD_FETCH_BOOKMARK, "bookmark"} };
gvals getdataextensions[] = { {SQL_GD_ANY_COLUMN, "anycolumn"},
  {SQL_GD_ANY_ORDER, "anyorder"},
  {SQL_GD_BLOCK, "block"},
  {SQL_GD_BOUND, "bound"} };
gvals locktypes[] = { {SQL_LCK_NO_CHANGE, "nochange"},
  {SQL_LCK_EXCLUSIVE, "exclusive"},
  {SQL_LCK_UNLOCK, "unlock"} };
gvals posoperations[] = { {SQL_POS_POSITION, "position"},
  {SQL_POS_REFRESH, "refresh"},
  {SQL_POS_UPDATE, "update"},
  {SQL_POS_DELETE, "delete"},
  {SQL_POS_ADD, "add"} };
gvals fileusage[] = { {SQL_FILE_NOT_SUPPORTED, "notsupported"},
  {SQL_FILE_TABLE, "table"},
  {SQL_FILE_QUALIFIER, "qualifier"} };
gvals odbcapiconformance[] = { {SQL_OAC_NONE, "none"},
  {SQL_OAC_LEVEL1, "level1"},
  {SQL_OAC_LEVEL2, "level2"} };
gvals odbcsagcliconformance[] = { {SQL_OSCC_NOT_COMPLIANT, "notcompliant"},
  {SQL_OSCC_COMPLIANT, "compliant"} };
ginfo info_driver[] = { {SQL_ACTIVE_CONNECTIONS, INT16, "activeconnections"},
  {SQL_ACTIVE_STATEMENTS, INT16, "activestatements"},
  {SQL_DATA_SOURCE_NAME, STRING, "datasourcename"},
  {SQL_DRIVER_HDBC, INT32, "driverhdbc"},
  {SQL_DRIVER_HENV, INT32, "driverhenv"},
  {SQL_DRIVER_HLIB, INT32, "driverhlib"},
  {SQL_DRIVER_NAME, STRING, "drivername"},
  {SQL_DRIVER_ODBC_VER, STRING, "driverodbcver"},
  {SQL_DRIVER_VER, STRING, "driverver"},
  {SQL_FETCH_DIRECTION, MASK32, XLAT(fetchdirection)},
  {SQL_FILE_USAGE, ENUM16, XLAT(fileusage)},
  {SQL_GETDATA_EXTENSIONS, MASK32, XLAT(getdataextensions)},
  {SQL_LOCK_TYPES, MASK32, XLAT(locktypes)},
  {SQL_ODBC_API_CONFORMANCE, ENUM16, XLAT(odbcapiconformance)},
  {SQL_ODBC_SAG_CLI_CONFORMANCE, ENUM16, XLAT(odbcsagcliconformance)},
  {SQL_ODBC_VER, STRING, "odbcver"},
  {SQL_POS_OPERATIONS, MASK32, XLAT(posoperations)},
  {SQL_ROW_UPDATES, STRING, "rowupdates"},
  {SQL_SEARCH_PATTERN_ESCAPE, STRING, "searchpatternescape"},
  {SQL_SERVER_NAME, STRING, "servername"} };

gvals bookmarkpersistence[] = { {SQL_BP_CLOSE, "close"},
  {SQL_BP_DELETE, "delete"},
  {SQL_BP_DROP, "drop"},
  {SQL_BP_SCROLL, "scroll"},
  {SQL_BP_TRANSACTION, "transaction"},
  {SQL_BP_UPDATE, "update"},
  {SQL_BP_OTHER_HSTMT, "otherhstmt"} };
gvals concatnullbehavior[] = { {SQL_CB_NULL, "null"},
  {SQL_CB_NON_NULL, "nonnull"} };
gvals cursorcommitbehavior[] = { {SQL_CB_DELETE, "delete"},
  {SQL_CB_CLOSE, "close"},
  {SQL_CB_PRESERVE, "preserve"} };
gvals defaulttxnisolation[] = { {SQL_TXN_READ_UNCOMMITTED, "readuncommited"},
  {SQL_TXN_READ_COMMITTED, "readcommitted"},
  {SQL_TXN_REPEATABLE_READ, "repeatableread"},
  {SQL_TXN_SERIALIZABLE, "serializable"},
  {SQL_TXN_VERSIONING, "versioning"} };
gvals nullcollation[] = { {SQL_NC_END, "end"},
  {SQL_NC_HIGH, "high"},
  {SQL_NC_LOW, "low"},
  {SQL_NC_START, "start"} };
gvals scrollconcurrency[] = { {SQL_SCCO_READ_ONLY, "readonly"},
  {SQL_SCCO_LOCK, "lock"},
  {SQL_SCCO_OPT_ROWVER, "rowver"},
  {SQL_SCCO_OPT_VALUES, "values"} };
gvals scrolloptions[] = { {SQL_SO_FORWARD_ONLY, "forwardonly"},
  {SQL_SO_STATIC, "static"},
  {SQL_SO_KEYSET_DRIVEN, "keysetdriven"},
  {SQL_SO_DYNAMIC, "dynamic"},
  {SQL_SO_MIXED, "mixed"} };
gvals staticsensitivity[] = { {SQL_SS_ADDITIONS, "additions"},
  {SQL_SS_DELETIONS, "deletions"},
  {SQL_SS_UPDATES, "updates"} };
gvals txncapable[] = { {SQL_TC_NONE, "none"},
  {SQL_TC_DML, "dml"},
  {SQL_TC_DDL_COMMIT, "ddlcommit"},
  {SQL_TC_DDL_IGNORE, "ddlignore"},
  {SQL_TC_ALL, "all"} };

ginfo info_datasource[] = { {SQL_ACCESSIBLE_PROCEDURES, STRING, "accessibleprocedures"},
  {SQL_ACCESSIBLE_TABLES, STRING, "accessibletables"},
  {SQL_BOOKMARK_PERSISTENCE, MASK32, XLAT(bookmarkpersistence)},
  {SQL_CONCAT_NULL_BEHAVIOR, ENUM16, XLAT(concatnullbehavior)},
  {SQL_CURSOR_COMMIT_BEHAVIOR, ENUM16, XLAT(cursorcommitbehavior)},
  {SQL_CURSOR_ROLLBACK_BEHAVIOR, ENUM16, "cursorrollbackbehavior", cursorcommitbehavior,
              sizeof(cursorcommitbehavior) / sizeof(gvals)},
  {SQL_DATA_SOURCE_READ_ONLY, STRING, "datasourcereadonly"},
  {SQL_DEFAULT_TXN_ISOLATION, MASK32, XLAT(defaulttxnisolation)},
  {SQL_MULT_RESULT_SETS, STRING, "multresultsets"},
  {SQL_MULTIPLE_ACTIVE_TXN, STRING, "multipleactivetxn"},
  {SQL_NEED_LONG_DATA_LEN, STRING, "needlongdatalen"},
  {SQL_NULL_COLLATION, ENUM16, XLAT(nullcollation)},
  {SQL_OWNER_TERM, STRING, "ownerterm"},
  {SQL_PROCEDURE_TERM, STRING, "procedureterm"},
  {SQL_QUALIFIER_TERM, STRING, "qualifierterm"},
  {SQL_SCROLL_CONCURRENCY, MASK32, XLAT(scrollconcurrency)},
  {SQL_SCROLL_OPTIONS, MASK32, XLAT(scrolloptions)},
  {SQL_STATIC_SENSITIVITY, MASK32, XLAT(staticsensitivity)},
  {SQL_TABLE_TERM, STRING, "tableterm"},
  {SQL_TXN_CAPABLE, ENUM16, XLAT(txncapable)},
  {SQL_TXN_ISOLATION_OPTION, MASK32, "txnisolationoption", defaulttxnisolation,
              sizeof(defaulttxnisolation) / sizeof(gvals)},
  {SQL_USER_NAME, STRING, "username"} };

gvals convertfunctions[] = { {SQL_FN_CVT_CONVERT, "convert"} };
gvals numericfunctions[] = { {SQL_FN_NUM_ABS, "abs"},
  {SQL_FN_NUM_ACOS, "acos"},
  {SQL_FN_NUM_ASIN, "asin"},
  {SQL_FN_NUM_ATAN, "atan"},
  {SQL_FN_NUM_ATAN2, "atan2"},
  {SQL_FN_NUM_CEILING, "ceiling"},
  {SQL_FN_NUM_COS, "cos"},
  {SQL_FN_NUM_COT, "cot"},
  {SQL_FN_NUM_DEGREES, "degrees"},
  {SQL_FN_NUM_EXP, "exp"},
  {SQL_FN_NUM_FLOOR, "floor"},
  {SQL_FN_NUM_LOG, "log"},
  {SQL_FN_NUM_LOG10, "log10"},
  {SQL_FN_NUM_MOD, "mod"},
  {SQL_FN_NUM_PI, "pi"},
  {SQL_FN_NUM_POWER, "power"},
  {SQL_FN_NUM_RADIANS, "radians"},
  {SQL_FN_NUM_RAND, "rand"},
  {SQL_FN_NUM_ROUND, "round"},
  {SQL_FN_NUM_SIGN, "sign"},
  {SQL_FN_NUM_SIN, "sin"},
  {SQL_FN_NUM_SQRT, "sqrt"},
  {SQL_FN_NUM_TAN, "tan"},
  {SQL_FN_NUM_TRUNCATE, "truncate"} };
gvals stringfunctions[] = { {SQL_FN_STR_ASCII, "ascii"},
  {SQL_FN_STR_CHAR, "char"},
  {SQL_FN_STR_CONCAT, "concat"},
  {SQL_FN_STR_DIFFERENCE, "difference"},
  {SQL_FN_STR_INSERT, "insert"},
  {SQL_FN_STR_LCASE, "lcase"},
  {SQL_FN_STR_LEFT, "left"},
  {SQL_FN_STR_LENGTH, "length"},
  {SQL_FN_STR_LOCATE, "locate"},
  {SQL_FN_STR_LOCATE_2, "locate_2"},
  {SQL_FN_STR_LTRIM, "ltrim"},
  {SQL_FN_STR_REPEAT, "repeat"},
  {SQL_FN_STR_REPLACE, "replace"},
  {SQL_FN_STR_RIGHT, "right"},
  {SQL_FN_STR_RTRIM, "rtrim"},
  {SQL_FN_STR_SOUNDEX, "soundex"},
  {SQL_FN_STR_SPACE, "space"},
  {SQL_FN_STR_SUBSTRING, "substring"},
  {SQL_FN_STR_UCASE, "ucase"} };
gvals systemfunctions[] = { {SQL_FN_SYS_DBNAME, "dbname"},
  {SQL_FN_SYS_IFNULL, "ifnull"},
  {SQL_FN_SYS_USERNAME, "username"} };
gvals timedateintervals[] = { {SQL_FN_TSI_FRAC_SECOND, "frac_second"},
  {SQL_FN_TSI_SECOND, "second"},
  {SQL_FN_TSI_MINUTE, "minute"},
  {SQL_FN_TSI_HOUR, "hour"},
  {SQL_FN_TSI_DAY, "day"},
  {SQL_FN_TSI_WEEK, "week"},
  {SQL_FN_TSI_MONTH, "month"},
  {SQL_FN_TSI_QUARTER, "quarter"},
  {SQL_FN_TSI_YEAR, "year"} };
gvals timedatefunctions[] = { {SQL_FN_TD_CURDATE, "curdate"},
  {SQL_FN_TD_CURTIME, "curtime"},
  {SQL_FN_TD_DAYNAME, "dayname"},
  {SQL_FN_TD_DAYOFMONTH, "dayofmonth"},
  {SQL_FN_TD_DAYOFWEEK, "dayofweek"},
  {SQL_FN_TD_DAYOFYEAR, "dayofyear"},
  {SQL_FN_TD_HOUR, "hour"},
  {SQL_FN_TD_MINUTE, "minute"},
  {SQL_FN_TD_MONTH, "month"},
  {SQL_FN_TD_MONTHNAME, "monthname"},
  {SQL_FN_TD_NOW, "now"},
  {SQL_FN_TD_QUARTER, "quarter"},
  {SQL_FN_TD_SECOND, "second"},
  {SQL_FN_TD_TIMESTAMPADD, "timestampadd"},
  {SQL_FN_TD_TIMESTAMPDIFF, "timestampdiff"},
  {SQL_FN_TD_WEEK, "week"},
  {SQL_FN_TD_YEAR, "year"} };

ginfo info_scalarfunctions[] = { {SQL_CONVERT_FUNCTIONS, MASK32, XLAT(convertfunctions)},
  {SQL_NUMERIC_FUNCTIONS, MASK32, XLAT(numericfunctions)},
  {SQL_STRING_FUNCTIONS, MASK32, XLAT(stringfunctions)},
  {SQL_SYSTEM_FUNCTIONS, MASK32, XLAT(systemfunctions)},
  {SQL_TIMEDATE_ADD_INTERVALS, MASK32, "timedateaddintervals", timedateintervals,
              sizeof(timedateintervals) / sizeof(gvals)},
  {SQL_TIMEDATE_DIFF_INTERVALS, MASK32, "timedatediffintervals", timedateintervals,
              sizeof(timedateintervals) / sizeof(gvals)},
  {SQL_TIMEDATE_FUNCTIONS, MASK32, XLAT(timedatefunctions)} };

gvals conversions[] = { {SQL_CVT_BIGINT, "bigint"},
  {SQL_CVT_BINARY, "binary"},
  {SQL_CVT_BIT, "bit"},
  {SQL_CVT_CHAR, "char"},
  {SQL_CVT_DATE, "date"},
  {SQL_CVT_DECIMAL, "decimal"},
  {SQL_CVT_DOUBLE, "double"},
  {SQL_CVT_FLOAT, "float"},
  {SQL_CVT_INTEGER, "integer"},
  {SQL_CVT_LONGVARBINARY, "longvarbinary"},
  {SQL_CVT_LONGVARCHAR, "longvarchar"},
  {SQL_CVT_NUMERIC, "numeric"},
  {SQL_CVT_REAL, "real"},
  {SQL_CVT_SMALLINT, "smallint"},
  {SQL_CVT_TIME, "time"},
  {SQL_CVT_TIMESTAMP, "timestamp"},
  {SQL_CVT_TINYINT, "tinyint"},
  {SQL_CVT_VARBINARY, "varbinary"},
  {SQL_CVT_VARCHAR, "varchar"} };

#define CVT(id) MASK32, #id, conversions, sizeof(conversions) / sizeof(gvals)
ginfo info_conversion[] = { {SQL_CONVERT_BIGINT, CVT(bigint)},
  {SQL_CONVERT_BINARY, CVT(binary)},
  {SQL_CONVERT_BIT, CVT(bit)},
  {SQL_CONVERT_CHAR, CVT(char)},
  {SQL_CONVERT_DATE, CVT(date)},
  {SQL_CONVERT_DECIMAL, CVT(decimal)},
  {SQL_CONVERT_DOUBLE, CVT(double)},
  {SQL_CONVERT_FLOAT, CVT(float)},
  {SQL_CONVERT_INTEGER, CVT(integer)},
  {SQL_CONVERT_LONGVARBINARY, CVT(longvarbinary)},
  {SQL_CONVERT_LONGVARCHAR, CVT(longvarchar)},
  {SQL_CONVERT_NUMERIC, CVT(numeric)},
  {SQL_CONVERT_REAL, CVT(real)},
  {SQL_CONVERT_SMALLINT, CVT(smallint)},
  {SQL_CONVERT_TIME, CVT(time)},
  {SQL_CONVERT_TIMESTAMP, CVT(timestamp)},
  {SQL_CONVERT_TINYINT, CVT(tinyint)},
  {SQL_CONVERT_VARBINARY, CVT(varbinary)},
  {SQL_CONVERT_VARCHAR, CVT(varchar)} };

gvals altertable[] = { {SQL_AT_ADD_COLUMN, "addcolumn"},
  {SQL_AT_DROP_COLUMN, "dropcolumn"} };
gvals correlationname[] = { {SQL_CN_NONE, "none"},
  {SQL_CN_DIFFERENT, "different"},
  {SQL_CN_ANY, "any"} };
gvals groupby[] = { {SQL_GB_NOT_SUPPORTED, "notsupported"},
  {SQL_GB_GROUP_BY_EQUALS_SELECT, "equalsselect"},
  {SQL_GB_GROUP_BY_CONTAINS_SELECT, "containsselect"},
  {SQL_GB_NO_RELATION, "norelation"} };
gvals identifiercase[] = { {SQL_IC_UPPER, "upper"},
  {SQL_IC_LOWER, "lower"},
  {SQL_IC_SENSITIVE, "sensitive"},
  {SQL_IC_MIXED, "mixed"} };
gvals nonnullablecolumns[] = { {SQL_NNC_NULL, "null"},
  {SQL_NNC_NON_NULL, "nonnull"} };
gvals odbcsqlconformance[] = { {SQL_OSC_MINIMUM, "minimum"},
  {SQL_OSC_CORE, "core"},
  {SQL_OSC_EXTENDED, "extended"} };
gvals ownerusage[] = { {SQL_OU_DML_STATEMENTS, "dmlstatements"},
  {SQL_OU_PROCEDURE_INVOCATION, "procedureinvocation"},
  {SQL_OU_TABLE_DEFINITION, "tabledefinition"},
  {SQL_OU_INDEX_DEFINITION, "indexdefinition"},
  {SQL_OU_PRIVILEGE_DEFINITION, "privilegedefinition"} };
gvals positionedstatements[] = { {SQL_PS_POSITIONED_DELETE, "positioneddelete"},
  {SQL_PS_POSITIONED_UPDATE, "positionedupdate"},
  {SQL_PS_SELECT_FOR_UPDATE, "selectforupdate"} };
gvals qualifierlocation[] = { {SQL_QL_START, "start"},
  {SQL_QL_END, "end"} };
gvals qualifierusage[] = { {SQL_QU_DML_STATEMENTS, "dmlstatements"},
  {SQL_QU_PROCEDURE_INVOCATION, "procedureinvocation"},
  {SQL_QU_TABLE_DEFINITION, "tabledefinition"},
  {SQL_QU_INDEX_DEFINITION, "indexdefinition"},
  {SQL_QU_PRIVILEGE_DEFINITION, "privilegedefinition"} };
gvals subqueries[] = { {SQL_SQ_CORRELATED_SUBQUERIES, "correlatedsubqueries"},
  {SQL_SQ_COMPARISON, "comparison"},
  {SQL_SQ_EXISTS, "exists"},
  {SQL_SQ_IN, "in"},
  {SQL_SQ_QUANTIFIED, "quantified"} };
gvals sqlunion[] = { {SQL_U_UNION, "union"},
  {SQL_U_UNION_ALL, "unionall"} };

ginfo info_sql[] = { {SQL_ALTER_TABLE, MASK32, XLAT(altertable)},
  {SQL_COLUMN_ALIAS, STRING, "columnalias"},
  {SQL_CORRELATION_NAME, ENUM16, XLAT(correlationname)},
  {SQL_EXPRESSIONS_IN_ORDERBY, STRING, "expressionsinorderby"},
  {SQL_GROUP_BY, ENUM16, XLAT(groupby)},
  {SQL_IDENTIFIER_CASE, ENUM16, XLAT(identifiercase)},
  {SQL_IDENTIFIER_QUOTE_CHAR, STRING, "identifierquotechar"},
  {SQL_KEYWORDS, STRING, "keywords"},
  {SQL_LIKE_ESCAPE_CLAUSE, STRING, "likeescapeclause"},
  {SQL_NON_NULLABLE_COLUMNS, ENUM16, XLAT(nonnullablecolumns)},
  {SQL_ODBC_SQL_CONFORMANCE, ENUM16, XLAT(odbcsqlconformance)},
  {SQL_ODBC_SQL_OPT_IEF, STRING, "odbcsqloptief"},
  {SQL_ORDER_BY_COLUMNS_IN_SELECT, STRING, "orderbycolumnsinselect"},
  {SQL_OUTER_JOINS, STRING, "outerjoins"},
  {SQL_OWNER_USAGE, MASK32, XLAT(ownerusage)},
  {SQL_POSITIONED_STATEMENTS, MASK32, XLAT(positionedstatements)},
  {SQL_PROCEDURES, STRING, "procedures"},
  {SQL_QUALIFIER_LOCATION, ENUM16, XLAT(qualifierlocation)},
  {SQL_QUALIFIER_NAME_SEPARATOR, STRING, "qualifiernameseparator"},
  {SQL_QUALIFIER_USAGE, MASK32, XLAT(qualifierusage)},
  {SQL_QUOTED_IDENTIFIER_CASE, ENUM16, "quotedidentifiercase", identifiercase,
         sizeof(identifiercase) / sizeof(gvals)},
  {SQL_SPECIAL_CHARACTERS, STRING, "specialcharacters"},
  {SQL_SUBQUERIES, MASK32, XLAT(subqueries)},
  {SQL_UNION, MASK32, "union", sqlunion, sizeof(sqlunion) / sizeof(gvals)} };

ginfo info_limits[] = { {SQL_MAX_BINARY_LITERAL_LEN, INT32, "maxbinaryliterallen"},
  {SQL_MAX_CHAR_LITERAL_LEN, INT32, "maxcharliterallen"},
  {SQL_MAX_COLUMN_NAME_LEN, INT16, "maxcolumnnamelen"},
  {SQL_MAX_COLUMNS_IN_GROUP_BY, INT16, "maxcolumnsingroupby"},
  {SQL_MAX_COLUMNS_IN_ORDER_BY, INT16, "maxcolumnsinorderby"},
  {SQL_MAX_COLUMNS_IN_INDEX, INT16, "maxcolumnsinindex"},
  {SQL_MAX_COLUMNS_IN_SELECT, INT16, "maxcolumnsinselect"},
  {SQL_MAX_COLUMNS_IN_TABLE, INT16, "maxcolumnsintable"},
  {SQL_MAX_CURSOR_NAME_LEN, INT16, "maxcursornamelen"},
  {SQL_MAX_INDEX_SIZE, INT32, "indexsize"},
  {SQL_MAX_OWNER_NAME_LEN, INT16, "maxownernamelen"},
  {SQL_MAX_PROCEDURE_NAME_LEN, INT16, "maxprocedurenamelen"},
  {SQL_MAX_QUALIFIER_NAME_LEN, INT16, "maxqualifiernamelen"},
  {SQL_MAX_ROW_SIZE, INT32, "maxrowsize"},
  {SQL_MAX_ROW_SIZE_INCLUDES_LONG, STRING, "maxrowsizeincludeslong"},
  {SQL_MAX_STATEMENT_LEN, INT32, "maxstatementlen"},
  {SQL_MAX_TABLE_NAME_LEN, INT16, "maxtablenamelen"},
  {SQL_MAX_TABLES_IN_SELECT, INT16, "maxtablesinselect"},
  {SQL_MAX_USER_NAME_LEN, INT16, "maxusernamelen"} };

struct ginfo_list {
  ginfo *category;
  int items;
  char *label;
};
ginfo_list info_list[] = { {info_driver, sizeof(info_driver) / sizeof(ginfo), "driver"},
 {info_dbms, sizeof(info_dbms) / sizeof(ginfo), "dbms"},
 {info_datasource, sizeof(info_datasource) / sizeof(ginfo), "datasource"},
 {info_sql, sizeof(info_sql) / sizeof(ginfo), "sql"},
 {info_limits, sizeof(info_limits) / sizeof(ginfo), "limits"},
 {info_scalarfunctions, sizeof(info_scalarfunctions) / sizeof(ginfo), "scalarfunctions"},
 {info_conversion, sizeof(info_conversion) / sizeof(ginfo), "conversion"} };

int getinfoError::explain(Tcl_Interp *interp) {
  Tcl_ResetResult(interp);
  Tcl_AppendResult(interp, "bad category \"", msg, "\", must be: ", info_list[0].label, NULL);
  for (int  i = 1; i < sizeof(info_list) / sizeof(ginfo_list); i++)
    Tcl_AppendResult(interp, " | ", info_list[i].label, NULL);
  return TCL_ERROR;
}

//
// getinfo: gets information about driver & data source by category
//
int getinfo(pstate st, Tcl_Interp *interp, char *category)
{
  ginfo * info = NULL; int items, i;
  for (i = 0; i < sizeof(info_list) / sizeof(ginfo_list); i++)
    if ( ! strcmp(info_list[i].label, category) ) {
      info = info_list[i].category;
      items = info_list[i].items;
      break;
    }
  if ( ! info ) throw(getinfoError(category));
  for (i = 0; i < items; i++, info++) {
    SWORD cbInfo; unsigned int value32 = 0; char value[4096]; int n;
    RETCODE rc = SQLGetInfo(st->hdbc, info->infotype,
         info->datatype == STRING ? (PTR) value : (PTR) &value32, sizeof(value), &cbInfo);
    VERIFY( NULLHSTMT );
    Tcl_AppendResult(interp, i ? " {" : "{", info->label, NULL);
    switch ( info->datatype ) {
      case MASK16:
      case MASK32: Tcl_AppendResult(interp, " {", NULL);
                   for (n = 0; n < info->nvalues; n++)
                     if ( value32 & info->values[n].value )
                       Tcl_AppendElement(interp, info->values[n].label);
                   Tcl_AppendResult(interp, "}", NULL);
                   break;
      case ENUM16:
      case ENUM32: for (n = 0; n < info->nvalues; n++)
                     if ( value32 == info->values[n].value ) {
                       Tcl_AppendElement(interp, info->values[n].label); break;
                     }
                   if (n == info->nvalues) Tcl_AppendElement(interp, "unknwon");
                   break;
      case INT32:
      case INT16: wsprintf(value, "0x%x", value32);
      case STRING: Tcl_AppendElement(interp, value); break;
    }
    Tcl_AppendResult(interp, "}", NULL);
  }
  return TCL_OK;
}

//
// odbc_command: processes commands for a connection
//
int odbc_command(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  try {
    if ( argc < 2 ) { throw(argsError("method ?params? ...")); }
    if ( ! strcmp(argv[1], "execdirect") ) {
      int reset = bind_params(STATE, argc - 3, argv + 3);
      RETCODE rc = SQLExecDirect(STATE->hstmt, (UCHAR *) argv[2], SQL_NTS);
      VERIFY( reset ? RESETP : 0 );
      SQLFreeStmt(STATE->hstmt, SQL_RESET_PARAMS);
      get_results(STATE, interp);
    } else if ( ! strcmp(argv[1], "commit") ) {
      RETCODE rc = SQLTransact(henv, STATE->hdbc, SQL_COMMIT);
      VERIFY( NULLHSTMT );
    } else if ( ! strcmp(argv[1], "rollback") ) {
      RETCODE rc = SQLTransact(henv, STATE->hdbc, SQL_ROLLBACK);
      VERIFY( NULLHSTMT );
    } else if ( ! strcmp(argv[1], "cget") ) {
      if ( argc != 3 ) { throw(argsError("cget option")); }
      char value[SQL_MAX_OPTION_STRING_LENGTH+1];
      cget(STATE, option_identifier(argv[2]), value);
      Tcl_SetResult(interp, value, TCL_VOLATILE);
    } else if ( ! strcmp(argv[1], "configure") ) {
      config(STATE, interp, argc - 2, argv + 2);
    } else if ( ! strcmp(argv[1], "getfunctions") ) {
      UWORD exists[100];
      RETCODE rc = SQLGetFunctions(STATE->hdbc, SQL_API_ALL_FUNCTIONS, exists);
      VERIFY( NULLHSTMT );
      for (int i = 0; i < 100; i++) if ( exists[i] ) {
        char buf[10]; if ( !api_fn[i] ) wsprintf(buf, "%d-unknown", i);
        Tcl_AppendElement(interp, api_fn[i] ? api_fn[i] : buf);
      }
    } else if ( ! strcmp(argv[1], "gettypeinfo") ) {
      RETCODE rc = SQLGetTypeInfo(STATE->hstmt, SQL_ALL_TYPES);
      VERIFY( CLOSE );
      get_results(STATE, interp);
    } else if ( ! strcmp(argv[1], "nativesql") ) {
      if ( argc != 3 ) { throw(argsError("nativesql sql-statement")); }
      char nativeSql[2048]; SDWORD cb;
      RETCODE rc = SQLNativeSql(STATE->hdbc, (UCHAR *) argv[2], SQL_NTS,
                     (UCHAR *) nativeSql, sizeof(nativeSql), &cb);
      VERIFY( NULLHSTMT );
      Tcl_SetResult(interp, nativeSql, TCL_VOLATILE);
    } else if ( ! strcmp(argv[1], "tables") ) {
      if ( argc != 6 ) { throw(argsError("tables qualifier owner name type")); }
      RETCODE rc;
      if ( (!argv[3][0] && !argv[4][0] && !strcmp(argv[2], "%")) ||
           (!argv[2][0] && !argv[4][0] && !strcmp(argv[3], "%")) ||
           (!argv[2][0] && !argv[3][0] && !argv[4][0] && !strcmp(argv[5], "%")) ) // enumeration
        rc = SQLTables(STATE->hstmt, (UCHAR *) argv[2], SQL_NTS,
          (UCHAR *) argv[3], SQL_NTS, (UCHAR *) argv[4], SQL_NTS, (UCHAR *) argv[5], SQL_NTS);
      else
        rc = SQLTables(STATE->hstmt, SPARAM(2), SPARAM(3), SPARAM(4), SPARAM(5));
      VERIFY( CLOSE );
      get_results(STATE, interp);
    } else if ( ! strcmp(argv[1], "tableprivileges") ) {
      if ( argc != 5 ) { throw(argsError("tableprivileges qualifier owner name")); }
      RETCODE rc = SQLTablePrivileges(STATE->hstmt, SPARAM(2), SPARAM(3), SPARAM(4));
      VERIFY( CLOSE );
      get_results(STATE, interp);
    } else if ( ! strcmp(argv[1], "columns") ) {
      if ( argc != 6 ) { throw(argsError("columns qualifier owner table-name column-name")); }
      RETCODE rc = SQLColumns(STATE->hstmt, SPARAM(2), SPARAM(3), SPARAM(4), SPARAM(5));
      VERIFY( CLOSE );
      get_results(STATE, interp);
    } else if ( ! strcmp(argv[1], "columnprivileges") ) {
      if ( argc != 6 ) { throw(argsError("columnprivileges qualifier owner table-name column-name")); }
      RETCODE rc = SQLColumnPrivileges(STATE->hstmt, SPARAM(2), SPARAM(3), SPARAM(4), SPARAM(5));
      VERIFY( CLOSE );
      get_results(STATE, interp);
    } else if ( ! strcmp(argv[1], "statistics") ) {
      if ( argc != 7 ) { throw(argsError("statistics qualifier owner table-name unique|all ensure|quick")); }
      RETCODE rc = SQLStatistics(STATE->hstmt, SPARAM(2), SPARAM(3), SPARAM(4),
                      UWORD(strcmp(argv[5], "unique") ? SQL_INDEX_ALL : SQL_INDEX_UNIQUE),
                      UWORD(strcmp(argv[6], "ensure") ? SQL_QUICK : SQL_ENSURE));
      VERIFY( CLOSE );
      get_results(STATE, interp);
    } else if ( ! strcmp(argv[1], "foreignkeys") ) {
      if ( argc != 8 ) { throw(argsError("foreignkeys qualifier owner table-name foreign-qualifier foreign-owner foreign-table")); }
      RETCODE rc = SQLForeignKeys(STATE->hstmt, SPARAM(2), SPARAM(3), SPARAM(4),
                      SPARAM(5), SPARAM(6), SPARAM(7));
      VERIFY( CLOSE );
      get_results(STATE, interp);
    } else if ( ! strcmp(argv[1], "primarykeys") ) {
      if ( argc != 5 ) { throw(argsError("primarykeys qualifier owner table-name")); }
      RETCODE rc = SQLPrimaryKeys(STATE->hstmt, SPARAM(2), SPARAM(3), SPARAM(4));
      VERIFY( CLOSE );
      get_results(STATE, interp);
    } else if ( ! strcmp(argv[1], "procedures") ) {
      if ( argc != 5 ) { throw(argsError("procedures qualifier owner procedure-name")); }
      RETCODE rc = SQLProcedures(STATE->hstmt, SPARAM(2), SPARAM(3), SPARAM(4));
      VERIFY( CLOSE );
      get_results(STATE, interp);
    } else if ( ! strcmp(argv[1], "procedurecolumns") ) {
      if ( argc != 6 ) { throw(argsError("procedurecolumns qualifier owner procedure-name column-name")); }
      RETCODE rc = SQLProcedureColumns(STATE->hstmt, SPARAM(2), SPARAM(3), SPARAM(4), SPARAM(5));
      VERIFY( CLOSE );
      get_results(STATE, interp);
    } else if ( ! strcmp(argv[1], "specialcolumns") ) {
      if ( argc != 8 ) { throw(argsError("specialcolumns bestrowid|rowver qualifier owner table-name "\
             "currow|transaction|session nonulls|nullable")); }
      RETCODE rc = SQLSpecialColumns(STATE->hstmt,
                UWORD(strcmp(argv[2], "rowver") ? SQL_ROWVER : SQL_BEST_ROWID),
                SPARAM(3), SPARAM(4), SPARAM(5),
                UWORD(!strcmp(argv[6], "currow") ? SQL_SCOPE_CURROW :
                     !strcmp(argv[6], "transaction") ? SQL_SCOPE_TRANSACTION : SQL_SCOPE_SESSION),
                      UWORD(strcmp(argv[7], "nonulls") ? SQL_NULLABLE : SQL_NO_NULLS));
      VERIFY( CLOSE );
      get_results(STATE, interp);
    } else if ( ! strcmp(argv[1], "getinfo") ) {
      if ( argc != 3 ) { throw(argsError("getinfo category")); }
      getinfo(STATE, interp, argv[2]);
    } else {
      Tcl_AppendResult(interp, "unknown method: \"", argv[1], "\"", NULL);
      return TCL_ERROR;
    }
  }
  catch (sqlError& err) { return err.explain(interp, STATE); }
  catch (argsError& err) { return err.explain(interp, argv[0]); }
  catch (msgError& err) { return err.explain(interp); }
  catch (connOptsError& err) { return err.explain(interp); }
  catch (getinfoError& err) { return err.explain(interp); }
  return TCL_OK;
}

//
// odbc_disconnect is called when proc returned by odbc_connect is deleted
//
void odbc_disconnect(ClientData clientData)
{
  SQLFreeStmt(STATE->hstmt, SQL_DROP);
  SQLDisconnect(STATE->hdbc);
  SQLFreeConnect(STATE->hdbc);
  ckfree((char *) STATE);
}

//
// odbc_connect: receives 2 parameters:  'object' name, DSN-string
//               returns an odbc 'object'
//
int odbc_connect(ClientData, Tcl_Interp *interp, int argc, char *argv[])
{
  pstate st = (pstate) ckalloc(sizeof(state)); st->hstmt = NULL; st->result = def_result_var;
  try {
    if ( argc != 3 ) { throw(argsError("command-name DSN-string")); }
    SQLAllocConnect(henv, &st->hdbc);
    char buf[400]; SWORD cb;
    RETCODE rc = SQLDriverConnect(st->hdbc, NULL, (UCHAR *) argv[2], SQL_NTS,
	 (UCHAR *) buf, sizeof(buf), &cb, SQL_DRIVER_COMPLETE);
    VERIFY( NULLHSTMT );
    SQLAllocStmt(st->hdbc, &st->hstmt);
    Tcl_CreateCommand(interp, argv[1], odbc_command, st, odbc_disconnect);
    Tcl_SetResult(interp, argv[1], TCL_VOLATILE);
  }
  catch (sqlError& err) { return err.explain(interp, st); }
  catch (argsError& err) { return err.explain(interp, argv[0]); }
  return TCL_OK;
}

//
// odbc_datasources: retrieves a list of data sources. Each element is itself a list
//                   of two elements: DSN & description
//
int odbc_datasources(ClientData, Tcl_Interp *interp, int, char *[])
{
  try {
    char dsn[SQL_MAX_DSN_LENGTH + 1], description[1024]; SWORD cbDsn, cbDescription;
    RETCODE rc = SQLDataSources(henv, SQL_FETCH_FIRST, (UCHAR *) dsn, sizeof(dsn), &cbDsn,
                 (UCHAR *) description, sizeof(description), &cbDescription);
    while ( rc != SQL_NO_DATA_FOUND ) {
      VERIFY( NULLHSTMT );
      if (*interp->result)
        Tcl_AppendResult(interp, " {", NULL);
      else
        Tcl_AppendResult(interp, "{", NULL);
      Tcl_AppendElement(interp, dsn);
      Tcl_AppendElement(interp, description);
      Tcl_AppendResult(interp, "}", NULL);
      rc = SQLDataSources(henv, SQL_FETCH_NEXT, (UCHAR *) dsn, sizeof(dsn), &cbDsn,
                 (UCHAR *) description, sizeof(description), &cbDescription);
    }
  }
  catch (sqlError& err) { return err.explain(interp, NULL); }
  return TCL_OK;
}

//
// odbc_drivers: retrieves a list of drivers. Each elemnt is itself a list
//                   of two elements: driver description and a list of driver attributes.
//
int odbc_drivers(ClientData, Tcl_Interp *interp, int, char *[])
{
  try {
    char description[256], attributes[2048]; SWORD cbDescription, cbAttributes;
    RETCODE rc = SQLDrivers(henv, SQL_FETCH_FIRST, (UCHAR *) description,
                   sizeof(description), &cbDescription,
                   (UCHAR *) attributes, sizeof(attributes), &cbAttributes);
    while ( rc != SQL_NO_DATA_FOUND ) {
      VERIFY( NULLHSTMT );
      if (*interp->result)
        Tcl_AppendResult(interp, " {", NULL);
      else
        Tcl_AppendResult(interp, "{", NULL);
      Tcl_AppendElement(interp, description);
      Tcl_AppendResult(interp, " {", NULL);
      for (char *attr = attributes; strlen(attr); attr += strlen(attr) + 1) {
        Tcl_AppendElement(interp, attr);
      }
      Tcl_AppendResult(interp, "}}", NULL);
      rc = SQLDrivers(henv, SQL_FETCH_NEXT, (UCHAR *) description,
                   sizeof(description), &cbDescription,
                   (UCHAR *) attributes, sizeof(attributes), &cbAttributes);
    }
  }
  catch (sqlError& err) { return err.explain(interp, NULL); }
  return TCL_OK;
}

extern "C" _export int Odbctcl_Init(Tcl_Interp *interp)
{
  if ( ! henv ) { // avoid initializing twice when using multiple interps!
    SQLAllocEnv(&henv);
    initialize();
  }
  Tcl_CreateCommand(interp, "odbc_connect", odbc_connect, NULL, NULL);
  Tcl_CreateCommand(interp, "odbc_datasources", odbc_datasources, NULL, NULL);
  Tcl_CreateCommand(interp, "odbc_drivers", odbc_drivers, NULL, NULL);
  return TCL_OK;
}

